"""
Utility functions for genice3
"""

import numpy as np
from dataclasses import dataclass
import string
from logging import getLogger

import networkx as nx
from cycless import center_of_graph
from cycless.cycles import cycles_iter
from cycless.polyhed import cage_to_graph, polyhedra_iter
from graphstat import GraphStat


@dataclass
class CageSpec:
    """ケージの仕様。cage_type は "A12" などのケージ種別を表す文字列。"""

    cage_type: str  # A12, 1b, etc.
    faces: str  # 5^12 6^2, etc.
    graph: nx.Graph  # ケージを構成する水分子ノードのグラフ

    def to_json_capable_data(self) -> dict:
        return {
            "cage_type": self.cage_type,
            "faces": self.faces,
            "nodes": [int(x) for x in self.graph],
        }

    def __repr__(self) -> str:
        return (
            f"CageSpec(cage_type={self.cage_type!r}, "
            f"faces={self.faces!r}, "
            f"n_nodes={self.graph.number_of_nodes()})"
        )

    def __str__(self) -> str:
        return f"{self.cage_type} ({self.faces}) {self.graph.number_of_nodes()} nodes"


@dataclass
class CageSpecs:
    """拡大単位胞内の全ケージの位置と仕様。

    specs: 各ケージの CageSpec（cage_type, faces, graph）。
    positions: 分数座標の配列（Nx3）。specs と同順。
    node_to_cage_indices: __post_init__ で生成。ノード番号 → そのノードを含むケージのインデックスリスト。
    """

    specs: list[CageSpec]
    positions: np.ndarray  # in fractional coordinates

    def __post_init__(self) -> None:
        """ノード番号 → ケージ番号の逆引き辞書を生成。"""
        self.node_to_cage_indices: dict[int, list[int]] = {}
        for cage_idx, spec in enumerate(self.specs):
            for node in spec.graph:
                self.node_to_cage_indices.setdefault(int(node), []).append(cage_idx)

    def to_json_capable_data(self) -> dict:
        """JSON シリアライズ用の辞書を返す。キーはケージインデックス、値は frac_pos と specs。"""
        data = []
        for position, specs in zip(self.positions, self.specs):
            data.append(
                {
                    "frac_pos": position.tolist(),
                    "specs": specs.to_json_capable_data(),
                }
            )
        return dict(enumerate(data))

    def __repr__(self) -> str:
        return (
            f"CageSpecs(n_cages={len(self.specs)}, "
            f"positions_shape={self.positions.shape})"
        )


def _assign_cage_type(basename: int, existing_types: set[str]) -> str:
    """未使用のケージタイプ名（A12, A12a など）を生成する。"""
    enum = 0
    cage_type = f"A{basename}"
    while cage_type in existing_types:
        char = string.ascii_lowercase[enum]
        cage_type = f"A{basename}{char}"
        enum += 1
    return cage_type


def _make_cage_expression(ring_ids: list, ringlist: list) -> str:
    """ケージを構成するリングのサイズから "5^12 6^2" のような面表現文字列を生成する。"""
    ringcount = [0 for i in range(9)]
    for ring in ring_ids:
        ringcount[len(ringlist[ring])] += 1
    index = []
    for i in range(9):
        if ringcount[i] > 0:
            index.append(f"{i}^{ringcount[i]}")
    index = " ".join(index)
    return index


def assess_cages(
    graph: nx.Graph, node_frac: np.ndarray
) -> CageSpecs:
    """水素結合グラフとノードの分数座標からケージを検出・分類し、CageSpecs を返す。

    Args:
        graph: 水素結合ネットワーク（無向グラフ）。
        node_frac: ノードの分数座標（Nx3）。

    Returns:
        検出されたケージの位置（分数座標）と仕様（CageSpec のリスト）。
    """
    logger = getLogger()

    # Prepare the list of rings
    # taking the positions in PBC into account.
    ringlist = [
        [int(x) for x in ring]
        for ring in cycles_iter(nx.Graph(graph), 8, pos=node_frac)
    ]

    MaxCageSize = 22
    cage_fracs = []
    # data storage of the found cages
    db = GraphStat()
    existing_cage_types: set[str] = set()
    g_id_to_cage_type: dict[int, str] = {}

    # Detect cages and classify
    cages = [cage for cage in polyhedra_iter(ringlist, MaxCageSize)]
    cage_graphs = [cage_to_graph(cage, ringlist) for cage in cages]
    cage_fracs = [center_of_graph(g, node_frac) for g in cage_graphs]
    if len(cage_fracs) == 0:
        logger.info("    No cages detected.")

    cagespecs = []
    for cage, g in zip(cages, cage_graphs):
        cagesize = len(cage)
        graph_id = db.query_id(g)
        # if it is a new cage type
        if graph_id < 0:
            # new type!
            # register the last query
            graph_id = db.register()

            cage_type = _assign_cage_type(cagesize, existing_cage_types)
            g_id_to_cage_type[graph_id] = cage_type
            existing_cage_types.add(cage_type)
        else:
            cage_type = g_id_to_cage_type[graph_id]
        faces = _make_cage_expression(cage, ringlist)
        cagespecs.append(CageSpec(cage_type=cage_type, faces=faces, graph=g))
        # print(f"{label=}, {faces=}, {ringlist=}")
        # print([len(ringlist[ring]) for ring in cage])

    return CageSpecs(specs=cagespecs, positions=np.array(cage_fracs))



from dataclasses import dataclass
import numpy as np


@dataclass
class Group:
    """
    Base class of a group

    - 置換イオンと単結合を持つものに限る。
    - 置換イオンから直結する原子への結合はz軸正方向を向いている。
    - 置換イオンに直結する原子の位置を原点とする。
    """

    sites: np.ndarray  # 原子の位置
    labels: list[str]  # 原子のラベル
    bonds: list[tuple[int, int]]  # 結合のリスト
    name: str  # グループの名前



"""
genice3 CLI パッケージ。

- option_parser: option_handling.md に基づくオプション文字列パーサー
- runner: argv のパースとプラグイン dispatch（parse_argv, validate_result）
"""

from genice3.cli.option_parser import (
    parse_options,
    structure_for_display,
    scalarize_single_item_lists,
    structure_to_option_string,
)

__all__ = [
    "parse_options",
    "structure_for_display",
    "scalarize_single_item_lists",
    "structure_to_option_string",
]



"""
CLI 利用者目線のテスト（使いやすさ・エラーメッセージのわかりやすさ）

【準備すべきテストの種類】

1. ヘルプ・バージョン
   - --help / -h: 使い方が分かる文言が表示される
   - --version / -V: バージョンが表示される
   - 終了コード 0 であること

2. 必須引数不足
   - UNITCELL 未指定: 「unitcell名が指定されていません」など明確なメッセージが stderr に出る
   - 終了コード 1 であること
   - メッセージが日本語で分かりやすいこと（必要なら英語でも一貫性を検証）

3. 不正なオプション値
   - --rep に不正な値（例: 0, 負の数, 非数）を渡したときのメッセージ
   - --seed に非整数を渡したときのメッセージ
   - エラー時に「何が悪いか」「どう直すか」が分かる文言を含むこと

4. 存在しないプラグイン名
   - 存在しない unitcell 名を指定したとき（例: --exporter gromacs NoSuchUnitcell）
   - 存在しない exporter 名を指定したとき（例: A15 --exporter NoSuchExporter）
   - 「Nonexistent or failed to load」「Dubious ... name」など、ユーザーが次に何をすべきか分かるメッセージ

5. 成功時の振る舞い
   - 正常系（例: A15 --exporter gromacs --rep 1 1 1）で exit 0、stdout に何か出力される
   - エラー時は stdout にスタックトレースを出さない（本番では logger のみにし、stderr に短いメッセージを出す設計が望ましい）

6. エラー時の出力形式
   - パースエラー時: traceback を出さず、1行程度のメッセージ＋必要なら「--help で使い方を確認」などの案内
   - バリデーションエラー時: 複数エラーを列挙する場合、箇条書きや改行で読みやすいこと

このファイルでは、上記のうち 1・2 と正常系の一部を subprocess で実行して検証する。
それ以外は、改善実装に合わせてテストを追加する。
"""

import subprocess
import sys
from pathlib import Path

# プロジェクトルートをパスに追加
project_root = Path(__file__).resolve().parent.parent.parent
if str(project_root) not in sys.path:
    sys.path.insert(0, str(project_root))


def _run_genice3(args: list[str], timeout: int = 10):
    """genice3 を python -m で実行し、(returncode, stdout, stderr) を返す"""
    cmd = [sys.executable, "-m", "genice3.cli.genice"] + args
    result = subprocess.run(
        cmd,
        cwd=project_root,
        capture_output=True,
        text=True,
        timeout=timeout,
        env={**__import__("os").environ, "PYTHONPATH": str(project_root)},
    )
    return result.returncode, result.stdout, result.stderr


# ---------------------------------------------------------------------------
# 1. ヘルプ・バージョン
# ---------------------------------------------------------------------------


def test_help_exits_zero_and_shows_usage():
    """--help で終了コード 0 かつ Usage が表示される"""
    code, out, err = _run_genice3(["--help"])
    assert code == 0, f"Expected exit 0, got {code}. stderr: {err}"
    assert "Usage:" in out or "usage:" in out.lower(), f"Expected Usage in stdout. got: {out[:500]}"
    assert "UNITCELL" in out or "unitcell" in out.lower()


def test_short_help_exits_zero():
    """-h でも同様にヘルプが表示される"""
    code, out, err = _run_genice3(["-h"])
    assert code == 0
    assert "Usage:" in out or "usage:" in out.lower()


def test_version_exits_zero_and_shows_version():
    """--version で終了コード 0 かつバージョン表示"""
    code, out, err = _run_genice3(["--version"])
    assert code == 0, f"Expected exit 0, got {code}. stderr: {err}"
    assert "genice3" in out
    # 3.0 などバージョンらしき数字
    assert any(c.isdigit() for c in out), f"Version-like string expected in: {out}"


# ---------------------------------------------------------------------------
# 2. 必須引数不足（バリデーションエラー）
# ---------------------------------------------------------------------------


def test_missing_unitcell_exits_nonzero_and_message_on_stderr():
    """UNITCELL 未指定時は終了コード 1 かつ stderr に分かりやすいメッセージ"""
    code, out, err = _run_genice3(["--exporter", "gromacs"])
    assert code == 1, f"Expected exit 1 when unitcell missing, got {code}. stdout: {out}, stderr: {err}"
    # 利用者に分かる文言（現在の実装は「unitcell名が指定されていません」）
    assert "unitcell" in err.lower() or "unitcell" in out.lower(), (
        f"Error message should mention unitcell. stderr: {err}, stdout: {out}"
    )


def test_missing_unitcell_message_is_user_friendly():
    """UNITCELL 未指定時のメッセージは短く、何を指定すべきか分かる"""
    code, out, err = _run_genice3(["--exporter", "gromacs"])
    combined = (err + "\n" + out).lower()
    # 単に「unitcell」に言及があるだけでなく、指定されていないことが分かる
    assert "指定" in (err + out) or "required" in combined or "unitcell" in combined


# ---------------------------------------------------------------------------
# 5. 成功時の振る舞い（軽いスモークテスト）
# ---------------------------------------------------------------------------


def test_valid_minimal_command_exits_zero_and_produces_stdout():
    """最小限の正常コマンドで終了コード 0 かつ stdout に出力がある"""
    code, out, err = _run_genice3(["1h", "--exporter", "gromacs", "--rep", "1", "1", "1"])
    assert code == 0, f"Expected exit 0. stderr: {err}"
    assert len(out.strip()) > 0, "Normal run should produce stdout (e.g. gro file)"


# ---------------------------------------------------------------------------
# 今後の拡張用: 不正オプション・存在しないプラグイン
# ---------------------------------------------------------------------------
# 以下は、エラーメッセージを改善したあとで assert を書くとよい。
#
# def test_invalid_rep_value_shows_clear_message():
#     code, out, err = _run_genice3(["1h", "--rep", "0", "0", "0"])
#     assert code == 1
#     assert "rep" in (err + out).lower() or "replication" in (err + out).lower()
#
# def test_nonexistent_unitcell_shows_plugin_error():
#     code, out, err = _run_genice3(["NoSuchUnitcell123", "--exporter", "gromacs"])
#     assert code == 1
#     assert "NoSuchUnitcell123" in (err + out) or "not found" in (err + out).lower()
#



#!/usr/bin/python
# coding: utf-8

import numpy as np
from genice3.util import (
    atomdic,
    symmetry_operators,
    waters_and_pairs,
    density_in_g_cm3,
)
import genice3.unitcell
import networkx as nx
from cif2ice import cellvectors

desc = {
    "ref": {"VIII": "Kuhs 1998"},
    "brief": "Ice VIII, a hydrogen-ordered counterpart of ice VII.",
    "test": ({"options": "--depol=none"},),
}


class UnitCell(genice3.unitcell.UnitCell):
    SUPPORTS_ION_DOPING = False  # 水素秩序氷
    def __init__(self, **kwargs):
        atoms = """
        O 0   0.25       0.1071(12)
        D 0   0.4157(15) 0.1935(8)
        """

        # space group: I4_1 /a m d No. 141

        symops = """
             x,            y,            z
             -x,          1/2-y,          z
            3/4-y,        1/4+x,        3/4+z
            3/4+y,        3/4-x,        1/4+z
             -x,            y,            z
              x,          1/2-y,          z
            1/4+y,        1/4+x,        3/4+z
            1/4-y,        3/4-x,        1/4+z

             -x,           -y,           -z
              x,          1/2+y,         -z
            3/4+y,        1/4-x,        3/4-z
            3/4-y,        3/4+x,        1/4-z
              x,           -y,           -z
             -x,          1/2+y,         -z
            1/4-y,        1/4-x,        3/4-z
            1/4+y,        3/4+x,        1/4-z

        """.translate(
            {ord(","): ""}
        )

        # add +1/2, +1/2, +1/2
        lines = ""
        for line in symops.split("\n"):
            cols = line.split()
            if len(cols) == 3:
                line = " ".join([x + "+1/2" for x in cols]) + "\n"
            lines += line

        symops += lines

        a = 4.656 / 10.0  # nm
        b = a
        c = 6.775 / 10.0  # nm
        A = 90
        B = 90
        C = 90

        cell = cellvectors(a, b, c, A, B, C)

        # helper routines to make from CIF-like data
        atomd = atomdic(atoms)
        sops = symmetry_operators(symops)
        waters, fixed = waters_and_pairs(cell, atomd, sops, rep=(2, 2, 2))

        density = density_in_g_cm3(len(waters), cell)
        coord = "relative"

        super().__init__(
            cell=cell,
            lattice_sites=waters,
            density=density,
            coord=coord,
            graph=nx.Graph(fixed),
            fixed=nx.DiGraph(fixed),
            **kwargs,
        )



"""
[1]
    Abascal, J.L.F., Sanz, E., García Fernández, R., Vega, C., 2005. A potential model for the study of ices and amorphous water: TIP4P/Ice. The Journal of chemical physics 122(23), 234511. DOI: 10.1063/1.1931662

J. L. F. Abascal, E. Sanz, R. G. Fernández, and C. Vega, A potential model for the study of ices and amorphous water: TIP4P/Ice, J. Chem. Phys. 122 (2005) 234511.
"""

from math import pi, sin, cos
import genice3.molecule
import numpy as np

desc = {
    "ref": {"ice": "Abascal 2005"},
    "usage": "No options available.",
    "brief": "TIP4P/Ice water model.",
}


class Molecule(genice3.molecule.Molecule):
    """
    TIP4P/Iceモデルの水分子を定義するクラス。
    """

    def __init__(self):
        L1 = 0.9572 / 10
        L2 = 0.1577 / 10
        theta = 104.52 * pi / 180

        hy = L1 * sin(theta / 2)
        hz = L1 * cos(theta / 2)
        mz = L2
        sites = np.array(
            [[0.0, 0.0, 0.0], [0.0, hy, hz], [0.0, -hy, hz], [0.0, 0.0, mz]]
        )
        sites -= (sites[1] + sites[2] + sites[0] * 0) / 18

        labels = ["OW", "HW1", "HW2", "MW"]
        name = "ICE"
        is_water = True
        super().__init__(sites=sites, labels=labels, name=name, is_water=is_water)


if __name__ == "__main__":
    water = Molecule()
    print(np.array([16, 1, 1, 0]) @ water.sites)



"""
setup.py for genice3
symlinkを実際のファイルにコピーするカスタムビルドフックを含む
"""
import os
from setuptools import setup
from setuptools.command.build_py import build_py as _build_py


class BuildPy(_build_py):
    """symlinkを実際のファイルにコピーするカスタムビルドコマンド"""

    def copy_file(self, src, dst, preserve_mode=True, preserve_times=True, link=None, level=1):
        """ファイルをコピーする際に、symlinkの場合は実際のファイルをコピー"""
        # symlinkの場合は、ターゲットファイルをコピー
        if os.path.islink(src):
            # symlinkのターゲットを取得
            target = os.readlink(src)
            
            # 相対パスの場合は、srcのディレクトリからの相対パスとして解決
            if not os.path.isabs(target):
                src_dir = os.path.dirname(src)
                target_path = os.path.normpath(os.path.join(src_dir, target))
            else:
                target_path = target
            
            # ターゲットファイルが存在する場合のみコピー
            if os.path.exists(target_path):
                # ターゲットファイルをコピー（再帰的に呼び出し）
                return super().copy_file(
                    target_path, dst, preserve_mode, preserve_times, link, level
                )
            else:
                # ターゲットが見つからない場合は警告を出してスキップ
                print(f"Warning: symlink target not found: {src} -> {target_path}")
                return dst, False
        
        # 通常のファイルの場合は親クラスのメソッドを呼び出す
        return super().copy_file(src, dst, preserve_mode, preserve_times, link, level)


setup(cmdclass={"build_py": BuildPy})



# coding: utf-8
"""
Test for generating water positions from cage positions
"""


import genice2.lattices
from genice2 import FrankKasper
from genice2.genice import parse_cages
from genice2.cell import cellvectors


class Lattice(genice2.lattices.Lattice):
    def __init__(self):
        self.cages = """
        12    0.5000    0.2500    0.2500
        12    0.5000    0.5000    0.5000
        12    0.2500    0.5000    0.2500
        12    0.0000    0.2500    0.7500
        12    0.2500    0.0000    0.7500
        12    0.0000    0.5000    0.0000
        12    0.2500    0.2500    0.5000
        12    0.5000    0.0000    0.0000
        12    0.7500    0.5000    0.7500
        12    0.0000    0.0000    0.5000
        12    0.0000    0.7500    0.2500
        12    0.2500    0.7500    0.0000
        12    0.7500    0.2500    0.0000
        12    0.7500    0.0000    0.2500
        12    0.5000    0.7500    0.7500
        12    0.7500    0.7500    0.5000
        16    0.6250    0.1250    0.6250
        16    0.1250    0.6250    0.6250
        16    0.8750    0.3750    0.3750
        16    0.3750    0.8750    0.3750
        16    0.3750    0.3750    0.8750
        16    0.6250    0.6250    0.1250
        16    0.8750    0.8750    0.8750
        16    0.1250    0.1250    0.1250
        """

        self.cell = cellvectors(a=12.747893943706936,
                                b=12.747893943706936,
                                c=12.747893943706936)
        cagepos, cagetype = parse_cages(self.cages)
        self.lattice_sites = [w for w in FrankKasper.toWater(cagepos, self.cell)]
        self.coord = "relative"
        self.density = FrankKasper.estimate_density(
            self.lattice_sites, self.cell, 2.76)
        self.bondlen = 2.76 * 1.1



# ============================================================================
# New genice3.unitcell implementation (TODO: implement manually)
# ============================================================================

"""
Test for generating water positions from cage positions
"""

import genice3.unitcell
import numpy as np
from cif2ice import cellvectors


class UnitCell(genice3.unitcell.UnitCell):
    """
    __test4単位胞を定義するクラス。

    NOTE: This unitcell is not yet implemented.
    Please contact the maintainer or implement it manually.
    """

    def __init__(self, **kwargs):
        raise NotImplementedError(
            f"{self.__class__.__name__} is not yet implemented. "
            "This unitcell requires manual implementation. "
            "Please contact the maintainer or implement it manually. "
            f"Reason: lattice_sitesが定義されていないため"
        )


import genice3.molecule
import numpy as np
import math

desc = {
    "ref": {"TIP7P": "Zhao 2019"},
    "brief": "A seven-site water model.",
    "usage": "No options available.",
}


class Molecule(genice3.molecule.Molecule):
    """
    7サイトモデルの水分子を定義するクラス。
    """

    def __init__(self):
        oh = 0.04786 * 2  # nm
        ol = 0.041  # nm
        hangle = math.radians(104.52) / 2
        mangle = math.radians(109.47) / 2
        mass = 18
        ohz = oh * math.cos(hangle)
        ohy = oh * math.sin(hangle)
        olz = -ol * math.cos(mangle)
        olx = ol * math.sin(mangle)
        oz = -ohz * 2 / mass

        sites = np.array(
            [
                [0, 0, oz],
                [0, ohy, ohz + oz],
                [0, -ohy, ohz + oz],
                [0, ohy / 2, ohz / 2 + oz],
                [0, -ohy / 2, ohz / 2 + oz],
                [olx, 0, olz + oz],
                [-olx, 0, olz + oz],
            ]
        )  # nm, OHHMMLL

        labels = ["OW", "HW1", "HW2", "MW1", "MW2", "LW1", "LW2"]
        name = "SOL"
        is_water = True
        super().__init__(sites=sites, labels=labels, name=name, is_water=is_water)




#!/usr/bin/python

import numpy as np
from genice2 import CIF
from genice2.cell import cellvectors
import genice2.lattices
desc = {
    "ref": {
        "14": 'Salzmann, C. G., Radaelli, P., Hallbrucker, A. & Mayer, E. The Preparation and Structures of Hydrogen Ordered Phases of Ice. Science 311, 1758–1761 (2006).'},
    "brief": "Ice XIV. (EXPERIMENTAL)"}

assert False, "IT DOES NOT WORK."


class Lattice(genice2.lattices.Lattice):
    def __init__(self):
        atoms = """
        O1 0.0059(3) 0.2568(5) 0.1304(7) 1.53(2) 1.0000
        O2 0.6308(3) -0.0078(3) 0.2485(7) 1.53(2) 1.0000
        O3 0.2525(4) 0.8858(3) 0.0063(6) 1.53(2) 1.0000
        D5 0.5275(5) 0.8410(4) 0.4684(1) 2.03(1) 0.620(4)
        D6 0.0920(2) 0.2056(3) 0.2671(6) 2.03(1) 1.0000
        D9 0.7895(3) 0.9679(3) 0.8954(7) 2.03(1) 1.0000
        D11 0.7340(3) 0.4630(3) 0.3225(6) 2.03(1) 1.0000
        D12 0.4111(4) 0.5790(5) 0.3625(1) 2.03(1) 0.593(3)
        D15 0.8472(3) 0.3248(3) 0.4010(6) 2.03(1) 1.0000
        """
        # Minor hydrogen sites omitted.
        # D4 0.0557(7) 0.3284(7) 0.9845(1) 2.03(1) 0.407(3)
        # D13 0.9018(8) 0.1018(7) 0.8552(2) 2.03(1) 0.380(4)

        # space group: P2_1 2_1 2_1 No. 19
        symops = """
             x,            y,            z
            1/2+x,         -y,          1/2-z
            1/2-x,        1/2+y,         -z
             -x,          1/2-y,        1/2+z
        """
        # http://img.chem.ucl.ac.uk/sgp/large/019az3.htm
        #x,            y,            z
        #    1/2+x,        1/2-y,         -z
        #     -x,          1/2+y,        1/2-z
        #    1/2-x,         -y,          1/2+z

        a = 8.3499 / 10.0  # nm
        b = 8.1391 / 10.0  # nm
        c = 4.0825 / 10.0  # nm

        self.cell = cellvectors(a, b, c)

        # helper routines to make from CIF-like data
        atomd = CIF.atomdic(atoms)
        sops = CIF.symmetry_operators(symops)
        self.lattice_sites, self.fixed = CIF.waters_and_pairs(self.cell, atomd, sops)

        # set self.pairs in this way for hydrogen-ordered ices.
        self.pairs = self.fixed

        self.density = 18 * len(self.lattice_sites) / 6.022e23 / \
            (np.linalg.det(self.cell) * 1e-21)

        self.coord = "relative"



# ============================================================================
# New genice3.unitcell implementation (TODO: implement manually)
# ============================================================================

desc = {'ref': {'14': 'Salzmann, C. G., Radaelli, P., Hallbrucker, A. & Mayer, E. The Preparation and Structures of Hydrogen Ordered Phases of Ice. Science 311, 1758–1761 (2006).'}, 'brief': 'Ice XIV. (EXPERIMENTAL)'}

import genice3.unitcell
import numpy as np
from genice3.util import cellvectors


class UnitCell(genice3.unitcell.UnitCell):
    """
    __ice14単位胞を定義するクラス。

    NOTE: This unitcell is not yet implemented.
    Please contact the maintainer or implement it manually.
    """

    def __init__(self, **kwargs):
        raise NotImplementedError(
            f"{self.__class__.__name__} is not yet implemented. "
            "This unitcell requires manual implementation. "
            "Please contact the maintainer or implement it manually. "
            f"Reason: CIFパターンを使用しているため"
        )


#!/usr/bin/python
"""
Usage: genice3 2
"""

desc = {
    "ref": {"2atom": "Kamb 2003", "2cell": "Kamb 1964", "C1": "Londono 1988"},
    "brief": "Hydrogen-ordered ice II.",
}

import genice3.unitcell
import numpy as np
from logging import getLogger
from genice3.util import (
    atomdic,
    symmetry_operators,
    waters_and_pairs,
    density_in_g_cm3,
)
import networkx as nx
from cif2ice import cellvectors


class UnitCell(genice3.unitcell.UnitCell):
    """
    ice2単位胞を定義するクラス。
    """

    SUPPORTS_ION_DOPING = False  # 水素秩序氷

    def __init__(self, **kwargs):
        logger = getLogger()

        # Ref. 2atom
        atoms = """
        O1   0.2716    0.0259   -0.1471
        O2   0.4798    0.7571    0.3389
        D1   0.7284    0.4038    0.4034
        D2   0.1491    0.0412   -0.2023
        D3   0.7420    0.1978    0.3708
        D4   0.4232    0.1954   -0.0164
        """

        # Ref. 2cell
        # space group: R-3
        symops = """
          x,            y,            z
          z,            x,            y
          y,            z,            x

         -x,           -y,           -z
         -z,           -x,           -y
         -y,           -z,           -x
        """.replace(
            ",", " "
        )

        # Ref. 2cell
        a = 7.78 / 10.0  # nm
        A = 113.1

        cell = cellvectors(a, a, a, A, A, A)

        # helper routines to make from CIF-like data
        atomd = atomdic(atoms)
        sops = symmetry_operators(symops)
        waters, fixed = waters_and_pairs(cell, atomd, sops)

        # set pairs in this way for hydrogen-ordered ices.
        fixed = nx.DiGraph(fixed)
        pairs = nx.Graph(fixed)

        density = density_in_g_cm3(len(waters), cell)

        coord = "relative"
        super().__init__(
            cell=cell,
            lattice_sites=waters,
            graph=pairs,
            fixed=fixed,
            density=density,
            coord=coord,
        )



import numpy as np

# United-atom THF model with a dummy site

import genice3.molecule


class Molecule(genice3.molecule.Molecule):
    """
    ダミーサイトを持つ統合原子THFモデルを定義するクラス。
    """

    def __init__(self):
        sites = np.array(
            [
                [0.0, -0.119625, 0.0],
                [0.116284, -0.039705, 0.0],
                [0.076453, 0.107915, 0.0],
                [-0.076453, 0.107915, 0.0],
                [-0.116284, -0.039705, 0.0],
                [0.0, 0.0, 0.0],
            ]
        )

        # self.atoms_ = ["O", "C", "C", "C", "C", "."]
        labels = ["O", "CA", "CB", "CB", "CA", "CM"]
        name = "THF"
        is_water = False
        super().__init__(sites=sites, labels=labels, name=name, is_water=is_water)




#!/usr/bin/env python3
"""
Generate a (hydrogen-ordered) Aeroice nxFAU.

CLI example: genice3 xFAU --length 3
"""

from genice3.unitcell import ice1c as ic  # base topology
import genice3.unitcell
from cif2ice import cellvectors
from genice3.util import shortest_distance
import re
from logging import getLogger
import numpy as np
from collections import defaultdict
from math import acos, pi, sin, cos
import networkx as nx
import pairlist as pl
from typing import Any, Dict, Tuple

desc = {
    "ref": {"xFAU": "Matsui 2017"},
    "brief": "Aeroice xFAU.",
    "options": [
        {
            "name": "length",
            "help": "Number of hexagonal prism segments. 0=SOD, 1=FAU, 2+=aeroice.",
            "required": True,
            "example": 3,
        },
    ],
    "test": (
        {"options": "--length 0"},
        {"options": "--length 1"},
        {"options": "--length 2"},
        {"options": "--length 4"},
        {"options": "--length 8"},
    ),
}


def _scalar(v: Any) -> Any:
    if isinstance(v, (list, tuple)) and len(v) == 1:
        return v[0]
    return v


def parse_options(options: Dict[str, Any]) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    """
    xFAU 固有: length（旧 rep も受け付ける）。
    残りは基底 UnitCell.parse_options に渡す。
    """
    processed: Dict[str, Any] = {}
    unprocessed = dict(options)
    for key in ("length", "rep"):
        if key in unprocessed:
            processed["length"] = _scalar(unprocessed.pop(key))
            break
    base_processed, base_unprocessed = genice3.unitcell.UnitCell.parse_options(
        unprocessed
    )
    processed.update(base_processed)
    return processed, base_unprocessed
# FAU Decoration of a 4-network
# 読みこんだAR3Rの座標を、FAU構造における多面体vertexの位置とみなし、
# それらを連結するネットワークを六角柱で修飾して大きなネットワークを作る。


# Aeroiceの超格子の接点多面体で、きちんとorderするように設計する。
# そのためには、超格子のもととなるdiamond latticeを白黒二部グラフとし、
# 黒から白へ向けていつも多角柱を作るようにする。
# 黒、白の接点多面体の六員環はすべてhomodromicとするが、どっちむきになるかは角柱の長さによって変わる。角柱が奇数段であれば、白黒とも同じ向きになるが、偶数段だと反転する。tune_anglesはてきとうに角柱を回転してつじつまをあわしているが、これを廃止する必要があるな。まずは全部決定論的に作る。
# ちょっとまじめに考えないといけない感じ。模型を睨む。


def tune_angles(sixvecs, pivot):
    """
    Find the best origin of angles to make sum cos(3 th) largest
    """
    sixangles = []
    for i in range(len(sixvecs)):
        vec = sixvecs[i]
        cosine = sixvecs[0] @ vec
        if cosine > 1.0:
            cosine = 1.0
        angle = acos(cosine)
        sine = np.cross(sixvecs[0], vec)
        if sine @ pivot < 0:
            angle = -angle
        sixangles.append(angle)
    offset = 0
    while True:
        sum = 0.0
        dsum = 0.0
        for a in sixangles:
            sum += cos((a + offset) * 6)
            dsum += -sin((a + offset) * 6)
        doffset = dsum / 20.0
        if abs(doffset) < 1e-6:
            return offset
        offset += doffset


class decorate:
    def __init__(self, atoms, cell, graph, Ncyl):
        """
        Ncyl is the number of cylinders to be inserted (>=0)
        """
        # make netghbor list
        self.diamond_graph = graph
        self.diamond_vertices = atoms
        self.diamond_cell = cell
        self.Ncyl = Ncyl
        self.vertices = []
        self.fixed_edges = []
        for edge in graph.edges():
            self.hexagonal_piller(edge)

    def hexagonal_piller(self, edge):
        logger = getLogger()
        i, j = edge
        dij = self.diamond_vertices[j] - self.diamond_vertices[i]
        dij -= np.floor(dij + 0.5)
        dij = dij @ self.diamond_cell
        scale = np.linalg.norm(dij)
        # dijは六角軸方向の単位ベクター
        dij /= scale
        # restsはiの隣接ノードのうちjを除いたもの
        rests = list(self.diamond_graph.neighbors(i))
        rests.remove(j)
        logger.debug(f"Rests: {rests}")
        # print(f"Rests: {list(rests)}")
        # Regularize the dihedral angles
        # to point them 6-fold directions.
        # by adding an offset
        vecs = []
        for k in rests:
            vec = self.diamond_vertices[k] - self.diamond_vertices[i]
            vec -= np.floor(vec + 0.5)
            vec = vec @ self.diamond_cell
            # orthogonalize
            shadow = dij @ vec
            vec -= shadow * dij
            vec /= np.linalg.norm(vec)
            vecs.append(vec)
        # print(f"{vecs=}")
        # 向きを同じにする。
        if np.linalg.det(np.vstack([dij, vecs[0], vecs[1]])) < 0:
            vecs[0], vecs[1] = vecs[1], vecs[0]
        offset = pi / 6  # 30 degree
        x = vecs[0]
        z = dij
        y = np.cross(z, x)
        sixvecs = np.zeros((6, 3))
        for j in range(6):
            a = j * pi * 2 / 6 + offset
            sixvecs[j] = x * cos(a) + y * sin(a)
        # determine r
        # assume edge length is 1
        # the radius of the outer sphere of the polyhed is sqrt(3/2)
        L = (3 / 2) ** 0.5 * 2 + self.Ncyl
        r = 1 / L  # edge len = radius of cyl
        rp = (3 / 2) ** 0.5 / L  # = radius of polyhed
        #
        icell = np.linalg.inv(self.diamond_cell)
        a = self.diamond_vertices[i] @ self.diamond_cell
        s = ""
        for j in range(0, self.Ncyl + 1):
            vec0 = dij * (rp + j * r) * scale + a
            for vec in sixvecs:
                rpos = vec0 + vec * r * scale
                pos = rpos @ icell
                # 水分子の位置。
                self.vertices.append(pos)

            first = len(self.vertices) - 6
            # 六角の結合を追加する。
            if j % 2 == 0:
                for k in range(5):
                    self.fixed_edges.append((first + k, first + k + 1))
                self.fixed_edges.append((first + 5, first))
            else:
                for k in range(5):
                    self.fixed_edges.append((first + k + 1, first + k))
                self.fixed_edges.append((first, first + 5))
            # 縦方向の結合を追加する交互に。
            if j > 0:
                for k in range(6):
                    if k % 2 == 0:
                        self.fixed_edges.append((first + k, first + k - 6))
                    else:
                        self.fixed_edges.append((first + k - 6, first + k))


class UnitCell(genice3.unitcell.UnitCell):
    """
    Generate a (hydrogen-ordered) Aeroice nxFAU.
    Option: length (CLI: --length N). 0:SOD, 1:FAU, 2+: aeroice.
    """

    def __init__(self, **kwargs):
        logger = getLogger()
        length = kwargs.get("length")
        if length is None:
            raise ValueError("xFAU unitcell requires --length N (e.g. genice3 xFAU --length 3)")
        Ncyl = int(length)
        ice1c = ic.UnitCell()
        cell1c = ice1c.cell
        waters1c = ice1c.lattice_sites
        graph1c = ice1c.graph
        #
        # 0..3を黒、4..7を白とする。もともと二部グラフになっているようだ。
        #
        logger.info("Superlattice {0}xFAU".format(Ncyl))
        dec = decorate(waters1c, cell1c, graph1c, Ncyl)

        coord = "relative"
        cell = cellvectors(a=dec.diamond_cell[0, 0], b=dec.diamond_cell[1, 1], c=dec.diamond_cell[2, 2])
        lattice_sites = np.array(dec.vertices)
        lattice_sites -= np.floor(lattice_sites)

        # 自力で密度を推定する。
        bondlen = shortest_distance(lattice_sites, cell)
        # 結合の長さはどんなに短くても0.276 nm。
        cell *= 0.276 / bondlen
        
        graph = nx.Graph([(i, j) for i, j in pl.pairs_iter(lattice_sites, maxdist=0.3, cell=cell, distance=False)])

        logger.info("after pairs_iter")
        super().__init__(
            cell=cell,
            lattice_sites=lattice_sites,
            coord=coord,
            fixed=nx.DiGraph(dec.fixed_edges),
            graph=graph,
        )



#!/usr/bin/python
# coding: utf-8

import numpy as np
from genice3.util import (
    atomdic,
    symmetry_operators,
    waters_and_pairs,
    density_in_g_cm3,
)
import genice3.unitcell
import networkx as nx
from cif2ice import cellvectors

desc = {
    "ref": {"A": "Baez 1998"},
    "brief": "Hypothetical ice A.",
}


class UnitCell(genice3.unitcell.UnitCell):
    SUPPORTS_ION_DOPING = False  # 水素秩序氷
    def __init__(self, **kwargs):
        atoms = """
        O1 0.0995 0.1901 0.2835
        O2 0.3431 0.3469 0.7598
        O3 0.4055 0.0558 0.5013
        H1 -0.0149 0.1524 0.3658
        H2 0.0921 0.3365 0.2570
        H3 0.4441 0.3484 0.8630
        H4 0.3713 0.2326 0.6746
        H5 0.3045 0.1078 0.4119
        H6 0.3847 -0.0897 0.5194
        """

        # P4_1, No. 76
        symops = """
              x            y            z
             -x           -y          1/2+z
             -y            x          1/4+z
              y           -x          3/4+z
        """

        # in nm
        a, b, c = 0.6733, 0.6733, 0.7164

        cell = cellvectors(a, b, c)

        # helper routines to make from CIF-like data
        atomd = atomdic(atoms)
        sops = symmetry_operators(symops)
        waters, fixed = waters_and_pairs(cell, atomd, sops)

        density = density_in_g_cm3(len(waters), cell)

        coord = "relative"

        super().__init__(
            cell=cell,
            lattice_sites=waters,
            density=density,
            coord=coord,
            graph=nx.Graph(fixed),
            fixed=nx.DiGraph(fixed),
            **kwargs,
        )



"""
c軸方向に分極した強誘電体氷11の単位胞を、a軸方向に上向き(u)または下向き(d)に並べ、a軸方向に長い単位胞を生成する。
uとdを切りかえるために、間に自動的に遷移層が挿入される。

例えば、genice3 11alt[uuuddd] と指定すると、実際にはuとdの間に分極のない遷移層"_"をはさんだ_uuu_dddという形の結晶が生成する。
"""
import genice3.unitcell
import numpy as np
from cif2ice import cellvectors
from genice3.util import density_in_g_cm3
from logging import getLogger
import networkx as nx

desc = {
    "ref": {},
    "brief": "A layered ferroelectric Ice XI.",
    "test": ({"layers": "udud"},),
}


class UnitCell(genice3.unitcell.UnitCell):
    SUPPORTS_ION_DOPING = False  # 水素秩序氷
    def __init__(self, **kwargs):
        logger = getLogger()

        sites = np.array(
            [
                [1 / 2, 0, 1 / 8],
                [1 / 2, 1 / 3, 0],
                [0, 1 / 2, 1 / 8],
                [0, 5 / 6, 0],
                [1 / 2, 0, 1 / 2],
                [1 / 2, 1 / 3, 5 / 8],
                [0, 1 / 2, 1 / 2],
                [0, 5 / 6, 5 / 8],
            ]
        )
        # c軸方向に上向きまたは下向きに分極した単位胞をa軸方向に重ねる。

        if len(kwargs) == 0:
            _layers = "uu"
        elif "layers" in kwargs:
            _layers = kwargs["layers"]
        else:
            for k, v in kwargs.items():
                if v is not True:
                    raise ValueError(f"Unknown option: {k}={v}")
                _layers = k
                # u, d以外の文字を含んでいればValueError
                if any(c not in "ud" for c in _layers):
                    raise ValueError(f"Invalid layer string: {_layers}")

        # parse options
        last = _layers[-1]
        layers = ""
        for layer in _layers:
            if last != layer:
                if layer == "u":
                    layers += "/"
                else:
                    layers += "\\"
            layers += layer
            last = layer
        logger.info(f"{layers}; {len(layers)} layers along the 'a' axis.")

        N_layers = len(layers)

        assert N_layers > 1, "More than one layer must be specified."

        edges = []
        waters = []
        last = layers[-1]
        for i, layer in enumerate(layers):
            waters.append(sites + np.array([i, 0, 0]))
            offset = i * 8
            right = (i + 1) * 8
            if i + 1 == N_layers:
                right = 0
            if layer == "u":
                assert last in ("u", "/"), "Incompatible u layer."
                edges.append([offset + 0, offset + 1])
                edges.append([offset + 1, offset + 2])
                edges.append([offset + 2, offset + 3])
                edges.append([offset + 3, offset + 0])
                edges.append([offset + 1, right + 2])
                edges.append([right + 3, offset + 0])
                edges.append([offset + 0, offset + 4])
                edges.append([offset + 2, offset + 6])
                edges.append([offset + 4, offset + 7])
                edges.append([offset + 7, offset + 6])
                edges.append([offset + 6, offset + 5])
                edges.append([offset + 5, offset + 4])
                edges.append([offset + 4, right + 7])
                edges.append([right + 6, offset + 5])
                edges.append([offset + 5, offset + 1])
                edges.append([offset + 7, offset + 3])
            elif layer == "d":
                assert last in ("d", "\\"), "Incompatible d layer."
                edges.append([offset + 0, offset + 3])
                edges.append([offset + 3, offset + 2])
                edges.append([offset + 2, offset + 1])
                edges.append([offset + 1, offset + 0])
                edges.append([right + 2, offset + 1])
                edges.append([offset + 0, right + 3])
                edges.append([offset + 4, offset + 0])
                edges.append([offset + 6, offset + 2])
                edges.append([offset + 7, offset + 4])
                edges.append([offset + 6, offset + 7])
                edges.append([offset + 5, offset + 6])
                edges.append([offset + 4, offset + 5])
                edges.append([right + 7, offset + 4])
                edges.append([offset + 5, right + 6])
                edges.append([offset + 1, offset + 5])
                edges.append([offset + 3, offset + 7])
            elif layer == "/":
                assert last in ("d", "\\"), "Incompatible d2u layer."
                edges.append([offset + 0, offset + 3])
                edges.append([offset + 3, offset + 2])
                edges.append([offset + 2, offset + 1])
                edges.append([offset + 0, offset + 1])
                edges.append([offset + 1, right + 2])
                edges.append([right + 3, offset + 0])
                edges.append([offset + 4, offset + 0])
                edges.append([offset + 6, offset + 2])

                edges.append([offset + 5, offset + 4])
                edges.append([offset + 5, offset + 6])
                edges.append([offset + 6, offset + 7])
                edges.append([offset + 7, offset + 4])
                edges.append([offset + 4, right + 7])
                edges.append([right + 6, offset + 5])
                edges.append([offset + 1, offset + 5])
                edges.append([offset + 3, offset + 7])
            elif layer == "\\":
                assert last in ("u", "/"), "Incompatible u2d layer."
                edges.append([offset + 1, offset + 0])
                edges.append([offset + 1, offset + 2])
                edges.append([offset + 2, offset + 3])
                edges.append([offset + 3, offset + 0])
                edges.append([right + 2, offset + 1])
                edges.append([offset + 0, right + 3])
                edges.append([offset + 0, offset + 4])
                edges.append([offset + 2, offset + 6])

                edges.append([offset + 4, offset + 7])
                edges.append([offset + 7, offset + 6])
                edges.append([offset + 6, offset + 5])
                edges.append([offset + 4, offset + 5])
                edges.append([right + 7, offset + 4])
                edges.append([offset + 5, right + 6])
                edges.append([offset + 5, offset + 1])
                edges.append([offset + 7, offset + 3])
            last = layer

        waters = np.vstack(waters) / np.array([N_layers, 1, 1])
        N_waters = N_layers * 8
        logger.debug(waters.shape)

        a = 4.4923 / 10 * N_layers
        b = 7.7808 / 10
        c = 7.3358 / 10

        cell = cellvectors(a, b, c)

        density = density_in_g_cm3(N_waters, cell)

        coord = "relative"

        super().__init__(
            cell=cell,
            lattice_sites=waters,
            density=density,
            coord=coord,
            graph=nx.Graph(edges),
            fixed=nx.DiGraph(edges),
        )



# It is a dummy unitcell to load a cif file.
# usage: genice3 CIF[file=diamond.cif, site=C]


import genice3.unitcell
import numpy as np
from typing import Any, Dict, List, Tuple
import networkx as nx
from cif2ice import cellvectors, read_cif
from genice3.util import atoms_to_waters, shortest_distance
import re
from logging import getLogger

desc = {
    "ref": {},
    "brief": "Load a CIF file and create a unit cell.",
    "options": [
        {"name": "file", "help": "Path to CIF file.", "required": True, "example": "path/to/structure.cif"},
        {"name": "osite", "help": "O site label or regex.", "required": False, "example": "O"},
        {"name": "hsite", "help": "H site label or regex (omit to let GenIce3 place H).", "required": False, "example": None},
    ],
    "test": ({"options": "--file ../../cif/MEP.cif --osite T"},),
}


def _scalar(v: Any) -> Any:
    if isinstance(v, (list, tuple)) and len(v) == 1:
        return v[0]
    return v


def parse_options(options: Dict[str, Any]) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    """
    CIF プラグインのオプションを処理する（file, osite, hsite, water_model）。
    残りは基底 UnitCell.parse_options に渡し、共通オプション（shift, density 等）をマージする。
    """
    processed: Dict[str, Any] = {}
    unprocessed = dict(options)
    for key in ("file", "osite", "hsite", "water_model"):
        if key in options:
            processed[key] = _scalar(options[key])
            unprocessed.pop(key, None)
    base_processed, base_unprocessed = genice3.unitcell.UnitCell.parse_options(unprocessed)
    processed.update(base_processed)
    return processed, base_unprocessed


class UnitCell(genice3.unitcell.UnitCell):
    """
    cif単位胞を定義するクラス。
    """

    def __init__(self, **kwargs):
        logger = getLogger("CIF")
        file = kwargs.get("file")
        osite = kwargs.get("osite")
        hsite = kwargs.get("hsite")
        if file is None:
            raise ValueError("file is required")
        if osite is None:
            osite = "O"

        # download(URL, fNameIn)
        # cif2ice はファイルを開けない場合に sys.exit(0) を呼ぶため、
        # SystemExit(0) を補足して例外に変換（exit code 1 にする）
        try:
            atoms, box = read_cif.read_and_process(file, make_rect_box=False)
        except SystemExit as e:
            if e.code in (0, None):
                raise ValueError(f"Failed to open CIF file '{file}'") from e
            raise
        # pattern matching
        oatoms = np.array([a[1:] for a in atoms if re.match(osite, a[0])])
        logger.info(f"{osite=} {oatoms.shape=} {atoms}")
        cell = cellvectors(*box) / 10  # nm
        shortest_OO = shortest_distance(oatoms, cell)
        cell *= 0.276 / shortest_OO

        # 共通 unitcell オプション（shift, density, anion, cation 等）は kwargs から親に渡す
        uc_kwargs = {
            k: v for k, v in kwargs.items()
            if k in ("shift", "density", "anion", "cation", "cation_groups")
        }
        if hsite is None:
            # 水素位置は指定されていないので、genice3にまかせる。
            super().__init__(
                cell=cell,  # nm
                lattice_sites=oatoms,
                coord="relative",
                **uc_kwargs,
            )
        else:
            hatoms = np.array([a[1:] for a in atoms if re.match(hsite, a[0])])
            waters, pairs, oo_pairs = atoms_to_waters(
                oatoms, hatoms, cell, partial_order=True
            )
            super().__init__(
                cell=cell,
                lattice_sites=waters,
                graph=nx.Graph(oo_pairs),
                fixed=nx.DiGraph(pairs),
                coord="relative",
                **uc_kwargs,
            )



import genice3.molecule
import numpy as np

desc = {
    "ref": {"H2": "https://www.britannica.com/science/hydrogen"},
    "brief": "Hydrogen molecule.",
    "usage": "No options available.",
}


class Molecule(genice3.molecule.Molecule):
    """
    水素分子を定義するクラス。
    """

    def __init__(self):
        sites = np.array([[0, 0, -0.037], [0, 0, +0.037]])  # nm, HH
        labels = ["H", "H"]
        name = "H2"
        is_water = False
        super().__init__(sites=sites, labels=labels, name=name, is_water=is_water)




#!/usr/bin/python
"""
Usage: genice3 c0te
"""

import genice3.unitcell
from genice3.util import (
    atomdic,
    fullatoms,
    symmetry_operators,
    waters_and_pairs,
    density_in_g_cm3,
)
from logging import getLogger
import numpy as np
from cif2ice import cellvectors


desc = {
    "ref": {"L": "Lei 2025"},
    "brief": "The hypothetical Ice L",
    "test": ({"options": "-r 2 2 2"},),
}


class UnitCell(genice3.unitcell.UnitCell):
    def __init__(self, **kwargs):
        atoms = """
        O1 0.3488 0.9861 0.3123
        O2 0.9710 0.6650 0.0774
        O3 0.3728 0.7575 0.1387
        """

        # space group: Pbca
        symops = """
        x,            y,            z
        1/2-x,        1/2+y,          z
        x,          1/2-y,        1/2+z
        1/2+x,          y,          1/2-z
    
        -x,           -y,           -z
        1/2+x,        1/2-y,         -z
        -x,          1/2+y,        1/2-z
        1/2-x,         -y,          1/2+z
        """.replace(
            ",", " "
        )

        # Ref. 2cell
        a = 6.23 / 10  # nm
        b = 7.24 / 10  # nm
        c = 11.88 / 10  # nm

        cell = cellvectors(a, b, c)

        atomd = atomdic(atoms)
        waters, _ = waters_and_pairs(cell, atomd, symmetry_operators(symops))

        density = density_in_g_cm3(len(waters), cell)

        coord = "relative"
        super().__init__(
            cell=cell,
            lattice_sites=waters,
            density=density,
            coord=coord,
            bondlen=0.3,
            **kwargs,
        )



import math
import numpy as np
import genice3.molecule

desc = {
    "ref": {"TIP3P": "Jorgensen 1983"},
    "usage": "No options available.",
    "brief": "A typical 3-site model.",
}


class Molecule(genice3.molecule.Molecule):
    """
    TIP3Pモデルの水分子を定義するクラス。
    """

    def __init__(self):
        oh = 0.09572
        hangle = 104.52 * math.pi / 180 / 2
        mass = 18
        ohz = oh * math.cos(hangle)
        ohy = oh * math.sin(hangle)
        oz = -ohz * 2 / mass
        sites = np.array(
            [[0, 0, oz], [0, ohy, ohz + oz], [0, -ohy, ohz + oz]]
        )  # nm, OHHM
        labels = ["O", "H", "H"]
        name = "SOL"
        is_water = True
        super().__init__(sites=sites, labels=labels, name=name, is_water=is_water)




# coding: utf-8
"""
Exporter that outputs cage positions and types as JSON.
Useful for understanding clathrate structures and reusing assessment data.
"""

import json
import sys
from collections import defaultdict
from io import TextIOWrapper
from logging import getLogger

from genice3.genice import GenIce3

format_desc = {
    "aliases": ["cage_survey"],
    "application": "JSON",
    "extension": ".json",
    "water": "none",
    "solute": "none",
    "hb": "none",
    "remarks": "Cage positions and types (fractional coords, labels, faces). Replaces -A/--assess_cages.",
}


def _brief_report(genice: GenIce3) -> None:
    """Log a brief summary of cage types and indices at INFO (for CLI)."""
    logger = getLogger()
    cages = genice.cages
    if cages is None or not cages.specs:
        return
    by_type: dict[str, list[int]] = defaultdict(list)
    for idx, spec in enumerate(cages.specs):
        by_type[spec.cage_type].append(idx)
    cage_types = sorted(by_type.keys())
    logger.info("Cage types: %s", cage_types)
    for ct in cage_types:
        indices = by_type[ct]
        logger.info("Cage type %s: %s", ct, set(indices))


def dumps(genice: GenIce3, **options) -> str:
    """Return cage survey as JSON string."""
    cages = genice.cages
    data = cages.to_json_capable_data()
    s = json.dumps(data, indent=4)
    return s if s.endswith("\n") else s + "\n"


def dump(genice: GenIce3, file: TextIOWrapper = sys.stdout, **options):
    """Write cage survey as JSON to file."""
    _brief_report(genice)
    file.write(dumps(genice, **options))



# coding: utf-8

from logging import getLogger
import numpy as np
from genice3.genice import GenIce3
from io import TextIOWrapper
import sys

desc = {
    "ref": {},
    "brief": "Calculate the polarization of the ice.",
    "usage": "Calculate the polarization of the ice.",
}


format_desc = {
    "aliases": ["_pol"],
    "application": "Polarization",
    "extension": "",
    "water": "",
    "solute": "",
    "hb": "none",
    "remarks": "Calculate the polarization of the ice.",
}


def calculate(genice: GenIce3) -> np.ndarray:
    "Calculate the polarization of the ice."
    logger = getLogger("genice3.exporter._pol")

    polarization = np.zeros(3)
    for i,j in genice.digraph.edges():
        pos_i = genice.lattice_sites[i]
        pos_j = genice.lattice_sites[j]
        delta = pos_j - pos_i
        delta -= np.floor(delta + 0.5)
        polarization += delta
    return polarization

def dumps(genice: GenIce3, **options) -> str:
    """
    Dump the polarization of the ice.
    """
    polarization = calculate(genice)
    return f"{polarization[0]:.5f} {polarization[1]:.5f} {polarization[2]:.5f} polarization\n"


def dump(genice: GenIce3, file: TextIOWrapper = sys.stdout, **options):
    file.write(dumps(genice, **options))
